 /***************************************************************************\
|*                                                                           *|
|*       Copyright 1993-2000 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NOTICE TO USER:   The source code  is copyrighted under  U.S. and     *|
|*     international laws.  Users and possessors of this source code are     *|
|*     hereby granted a nonexclusive,  royalty-free copyright license to     *|
|*     use this code in individual and commercial software.                  *|
|*                                                                           *|
|*     Any use of this source code must include,  in the user documenta-     *|
|*     tion and  internal comments to the code,  notices to the end user     *|
|*     as follows:                                                           *|
|*                                                                           *|
|*       Copyright 1993-2000 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NVIDIA, CORPORATION MAKES NO REPRESENTATION ABOUT THE SUITABILITY     *|
|*     OF  THIS SOURCE  CODE  FOR ANY PURPOSE.  IT IS  PROVIDED  "AS IS"     *|
|*     WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND.  NVIDIA, CORPOR-     *|
|*     ATION DISCLAIMS ALL WARRANTIES  WITH REGARD  TO THIS SOURCE CODE,     *|
|*     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, NONINFRINGE-     *|
|*     MENT,  AND FITNESS  FOR A PARTICULAR PURPOSE.   IN NO EVENT SHALL     *|
|*     NVIDIA, CORPORATION  BE LIABLE FOR ANY SPECIAL,  INDIRECT,  INCI-     *|
|*     DENTAL, OR CONSEQUENTIAL DAMAGES,  OR ANY DAMAGES  WHATSOEVER RE-     *|
|*     SULTING FROM LOSS OF USE,  DATA OR PROFITS,  WHETHER IN AN ACTION     *|
|*     OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,  ARISING OUT OF     *|
|*     OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOURCE CODE.     *|
|*                                                                           *|
|*     U.S. Government  End  Users.   This source code  is a "commercial     *|
|*     item,"  as that  term is  defined at  48 C.F.R. 2.101 (OCT 1995),     *|
|*     consisting  of "commercial  computer  software"  and  "commercial     *|
|*     computer  software  documentation,"  as such  terms  are  used in     *|
|*     48 C.F.R. 12.212 (SEPT 1995)  and is provided to the U.S. Govern-     *|
|*     ment only as  a commercial end item.   Consistent with  48 C.F.R.     *|
|*     12.212 and  48 C.F.R. 227.7202-1 through  227.7202-4 (JUNE 1995),     *|
|*     all U.S. Government End Users  acquire the source code  with only     *|
|*     those rights set forth herein.                                        *|
|*                                                                           *|
 \***************************************************************************/

#include <linux/module.h>

#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif

// pull in the versioned kernel api defines if needed
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <nv.h>			        // needs to precede other headers (SMP)

#include <linux/stddef.h>
#include <linux/mm.h>
#include <linux/kernel.h> 		// printk()
#include <linux/malloc.h> 		// kmalloc()
#include <linux/fs.h>     		// everything...
#include <linux/errno.h>  		// error codes
#include <linux/types.h>  		// size_t
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        	// O_ACCMODE
#include <linux/pci.h>    		// PCI stuff
#include <linux/delay.h>  		// udelay()
#include <linux/interrupt.h>           
#include <linux/tqueue.h>               // struct tq_struct 
#include <linux/poll.h>

#include <asm/page.h>
#include <asm/pgtable.h>  		// pte bit definitions
#include <asm/system.h>                 // cli(), *_flags
#include <asm/segment.h>                // memcpy and verify_write
#include <asm/uaccess.h>		// access_ok
#include <asm/io.h>
#include <asm/softirq.h>                // mark_bh
#ifdef KERNEL_2_3
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif
#include <asm/smp.h>

#include <linux/wrapper.h>

#include <nv_ref.h>

#define LINUX_VMA_DEV(vma)  ((vma)->vm_file->f_dentry->d_inode->i_rdev)

#ifdef KERNEL_2_3
#define LINUX_VMA_OFFS(vma) (((vma)->vm_pgoff) << PAGE_SHIFT)
#else
#define LINUX_VMA_OFFS(vma) ((vma)->vm_offset)
#endif

#define NV_PAGE_ALIGN(addr)    ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)  ( (addr) & (PAGE_SIZE - 1) )

/* 
 * Our main lock for everything.
 */

spinlock_t nv_lock = SPIN_LOCK_UNLOCKED;
unsigned long * volatile nv_locked;

/*
 * our global state; one per device
 */

nv_state_t nv_devices[NV_MAX_DEVICES];

/*
 * And one for the control device
 */

nv_state_t nv_ctl_device;

// keep track of opened clients and their process id so they
//   can be free'd up on abnormal close
nv_client_t       nv_clients[NV_MAX_CLIENTS];
struct tq_struct  nv_bottom_halves[NV_MAX_CLIENTS];

/*
 * pick apart our minor device number
 * low 3 bits is NV device
 * if 255, then its the control device
 */

#define NV_DEVICE_NUMBER(_minor) ((_minor) & 0x0f)
#define NV_DEVICE_IS_CONTROL_DEVICE(_minor) \
             (((_minor) & 0xFF) == 0xFF)

/*
 * Hardware access macros for the kernel driver only
 * The resource manager part of the driver uses different values here
 */

// HW register access macros
#define REG_WR08(a,d)   nv->regs_u->Reg008[(a)]   = (U008) (d)
#define REG_WR16(a,d)   nv->regs_u->Reg016[(a)/2] = (U016) (d)
#define REG_WR32(a,d)   nv->regs_u->Reg032[(a)/4] = (U032) (d)

#define REG_RD08(a)     nv->regs_u->Reg008[(a)]
#define REG_RD16(a)     nv->regs_u->Reg016[(a)/2]
#define REG_RD32(a)     nv->regs_u->Reg032[(a)/4]



#ifdef DEBUG
#define NV_DMSG(nv, fmt, args...)  NV_iDMSG(nv, fmt, ## args)
#else
#define NV_DMSG(nv, fmt, args...)
#endif


// forwards
int nv_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int nv_open(struct inode *, struct file *);
int nv_close(struct inode *, struct file *);
int nv_mmap(struct file *file, struct vm_area_struct *vma);
int nv_ctl_open(struct inode *, struct file *);
int nv_ctl_close(struct inode *, struct file *);
unsigned int nv_poll(struct file *file, poll_table *wait);

// character driver entry points

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
static struct file_operations nv_fops = {
    NULL,                   // lseek
    NULL,                   // read,
    NULL,                   // write
    NULL,                   // readdir
    nv_poll,
    nv_ioctl,
    nv_mmap,
    nv_open,
    NULL,                   // flush
    nv_close,		    // release
    NULL,                   // fsync
    NULL,                   // fasync
    NULL,                   // check_media_change
    NULL,                   // revalidate
    NULL,                   // lock
};
#else
static struct file_operations nv_fops = {
    owner:          THIS_MODULE,
    poll:           nv_poll,
    ioctl:          nv_ioctl,
    mmap:           nv_mmap, 
    open:           nv_open, 
    release:        nv_close,
};
#endif

// Our reserved major device number.
// Set this to 0 to request dynamic major number assignment
int nv_major = NV_MAJOR_DEVICE_NUMBER;


// do not call this directly
void NV_iMSG(
    nv_state_t *nv,
    const char *kernel_message_level,
    const char *printf_format,
    va_list arglist
  )
{
    char nv_error_string[256];
    char *p;

    p = nv_error_string;
    
    if (kernel_message_level)
        p += sprintf(p, "%s", kernel_message_level);

    if (nv)
        if (nv->device_number == NV_CONTROL_DEVICE_NUMBER)
            p += sprintf(p, "NVctl: ");
        else
            p += sprintf(p, "NV%d: ", nv->device_number);
    else
        p += sprintf(p, "NV: ");

    p += vsprintf(p, printf_format, arglist);
    printk("%s\n", nv_error_string);
}

// should only be called using NV_DMSG macro
void NV_iDMSG(
    nv_state_t *nv,
    const char *printf_format,
    ...
)
{
    va_list arglist;

    va_start(arglist, printf_format);
    NV_iMSG(nv, KERN_INFO, printf_format, arglist);
    va_end(arglist);
}

void NV_EMSG(
    nv_state_t *nv,
    const char *printf_format,
    ...
  )
{
    va_list arglist;

    va_start(arglist, printf_format);
    NV_iMSG(nv, KERN_ERR, printf_format, arglist);
    va_end(arglist);
}


//
// The following are the supported pci vendor_id/device_id pairs
//

#define NVIDIA_VENDOR 0x10DE

static unsigned long pciIds[] = {
    (0x20 << 16) + NVIDIA_VENDOR,
    (0x21 << 16) + NVIDIA_VENDOR,
    (0x22 << 16) + NVIDIA_VENDOR,
    (0x23 << 16) + NVIDIA_VENDOR,
    (0x24 << 16) + NVIDIA_VENDOR,
    (0x25 << 16) + NVIDIA_VENDOR,
    (0x26 << 16) + NVIDIA_VENDOR,
    (0x27 << 16) + NVIDIA_VENDOR,
    (0x28 << 16) + NVIDIA_VENDOR,
    (0x29 << 16) + NVIDIA_VENDOR,
    (0x2A << 16) + NVIDIA_VENDOR,
    (0x2B << 16) + NVIDIA_VENDOR,
    (0x2C << 16) + NVIDIA_VENDOR,
    (0x2D << 16) + NVIDIA_VENDOR,
    (0x2E << 16) + NVIDIA_VENDOR,
    (0x2F << 16) + NVIDIA_VENDOR,

    (0xA0 << 16) + NVIDIA_VENDOR,

    (0x100 << 16) + NVIDIA_VENDOR,
    (0x101 << 16) + NVIDIA_VENDOR,
    (0x102 << 16) + NVIDIA_VENDOR,
    (0x103 << 16) + NVIDIA_VENDOR,

    (0x110 << 16) + NVIDIA_VENDOR,		
    (0x111 << 16) + NVIDIA_VENDOR,		
    (0x112 << 16) + NVIDIA_VENDOR,		
    (0x113 << 16) + NVIDIA_VENDOR,		

    (0x150 << 16) + NVIDIA_VENDOR,
    (0x151 << 16) + NVIDIA_VENDOR,
    (0x152 << 16) + NVIDIA_VENDOR,
    (0x153 << 16) + NVIDIA_VENDOR,
    0
};

//
// Returns 1 if device is ours, 0 otherwise.
//

static int is_nv_device(
    unsigned char bus,
    unsigned char func,
    unsigned short vendor,
    unsigned short device
)
{
    unsigned long pciId = (device << 16) | vendor;
    unsigned long *id;

    NV_DMSG((nv_state_t *) 0,
            "is_nv_device: bus 0x%x fun 0x%x vendor 0x%x device 0x%x",
            bus, func, vendor, device);

    for (id = pciIds; *id != 0; id++)
        if (*id == pciId)
            return 1;

    return 0;
}

void nv_probe_devices(void)
{
    int probenum = 0;
    unsigned char bus, func;
    int dev_count;
    nv_state_t *nv;

    NV_DMSG((nv_state_t *) 0, "nv_probe_devices");

    //
    // Find nv devices and setup initial state.
    //
    dev_count = 0;
    while (PCIBIOS_SUCCESSFUL ==
           pcibios_find_class(PCI_CLASS_DISPLAY_VGA << 8, probenum,
                              &bus, &func))
    {
        unsigned short vendor_id, device_id;

        pcibios_read_config_word(bus, func, PCI_VENDOR_ID, &vendor_id);
        pcibios_read_config_word(bus, func, PCI_DEVICE_ID, &device_id);

        if (is_nv_device(bus, func, vendor_id, device_id))
        {
            unsigned short command_word;

            // Initialize bus-dependent config state.
            nv = &nv_devices[dev_count];
            nv->bus = bus;
            nv->slot = func >> 3;
            nv->vendor_id = vendor_id;
            nv->device_id = device_id;
            nv->device_number = dev_count;

            // is the device enabled currently?
            pcibios_read_config_word(bus, func, PCI_COMMAND, &command_word);
            if ( ! (command_word & PCI_COMMAND_MEMORY) ||
                 ! (command_word & PCI_COMMAND_MASTER))
            {
                NV_DMSG(nv, "device not PCI enabled; enabling");
                command_word |= PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
                pcibios_write_config_word(bus, func, PCI_COMMAND, command_word);
            }

#ifdef KERNEL_2_3
            nv->reg_physical_address = pci_find_slot(bus, func)->resource[0].start;
            nv->reg_physical_address &= PCI_BASE_ADDRESS_IO_MASK;

            nv->fb_physical_address = pci_find_slot(bus, func)->resource[1].start;
            nv->fb_physical_address &= PCI_BASE_ADDRESS_MEM_MASK;
            
            nv->interrupt_line = pci_find_slot(bus, func)->irq;
            /* XXX not sure of correct way to read interrupt_pin & 
             * we don't appear to use it anyways...
             */
#else
#if LINUX_VERSION_CODE >= 0x20155
            nv->reg_physical_address = pci_find_slot(bus, func)->base_address[0];
            nv->reg_physical_address &= PCI_BASE_ADDRESS_IO_MASK;

            nv->fb_physical_address = pci_find_slot(bus, func)->base_address[1];
            nv->fb_physical_address &= PCI_BASE_ADDRESS_MEM_MASK;

            nv->interrupt_line = pci_find_slot(bus, func)->irq;
            /* XXX not sure of correct way to read interrupt_pin & 
             * we don't appear to use it anyways...
             */
#else
            pcibios_read_config_dword(bus, func, PCI_BASE_ADDRESS_0, &addr);
            nv->reg_physical_address = addr & PCI_BASE_ADDRESS_IO_MASK;

            pcibios_read_config_dword(bus, func, PCI_BASE_ADDRESS_1, &addr);
            nv->fb_physical_address = addr & PCI_BASE_ADDRESS_MEM_MASK;

            pcibios_read_config_byte(bus, func, PCI_INTERRUPT_LINE, &intr);
            nv->interrupt_line = intr;

            pcibios_read_config_byte(bus, func, PCI_INTERRUPT_PIN, &intr);
            nv->interrupt_pin = intr;
#endif
#endif

            /* check common error condition */
            if (nv->interrupt_line == 0)
            {
                printk("NV: I can't find an interrupt for your nvidia card\n");
                printk("NV: you may need to change your bios settings:\n");
                printk("NV: plug and play OS should be set to 'NO'\n");
                printk("NV: VGA IRQ should be set to 'YES'\n");
            }

            NV_DMSG(nv,
                    "bus: %d,  slot: %d,  devid: 0x%x, reg phys: %08x,  fb phys: %08x",
                    nv->bus, nv->slot, nv->device_id,
                    (unsigned int) nv->reg_physical_address,
                    (unsigned int) nv->fb_physical_address);

            if (++dev_count == NV_MAX_DEVICES)
            {
                NV_EMSG(nv, "maximum devices (%d) exceeded!", NV_MAX_DEVICES);
                break;
            }
        }
        probenum++;
    }

    NV_DMSG((nv_state_t *) 0,
            "nv_probe_devices: found %d device%s", dev_count,
            dev_count == 1 ? "" : "s");
}


/* memory conversion routines */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 12)
static inline unsigned long 
get_kvirt_address(unsigned long address)
{
    pgd_t *pg_dir;
    pmd_t *pg_mid_dir;
    pte_t *pg_table;
    unsigned long ret_addr = 0;
    
    pg_dir = pgd_offset(current->mm, address);
    if (pgd_none(*pg_dir))
        goto done;

    pg_mid_dir = pmd_offset(pg_dir, address);
    if (pmd_none(*pg_mid_dir))
        goto done;

    pg_table = pte_offset(pg_mid_dir, address);
    if (!pte_present(*pg_table))
        goto done;

    ret_addr = pte_page(*pg_table) | NV_MASK_OFFSET(address);

  done:
    return ret_addr;
}

#else

static inline unsigned long 
get_kvirt_address(unsigned long address)
{
    pgd_t *pg_dir;
    pmd_t *pg_mid_dir;
    pte_t *pg_table;
    unsigned long ret_addr = 0;

    if (address > VMALLOC_START)
       pg_dir = pgd_offset_k(address);
    else
       pg_dir = pgd_offset(current->mm, address);

    if (pgd_none(*pg_dir))
        goto done;

    pg_mid_dir = pmd_offset(pg_dir, address);
    if (pmd_none(*pg_mid_dir))
        goto done;

    pg_table = pte_offset(pg_mid_dir, address);
    if (!pte_present(*pg_table))
        goto done;


#ifdef KERNEL_2_3
    ret_addr = ((unsigned long) __va(pte_val(*pg_table) & PAGE_MASK));
    ret_addr = (ret_addr | NV_MASK_OFFSET(address));
#else
    ret_addr = ((unsigned long)pte_page(*pg_table) | NV_MASK_OFFSET(address));
#endif

  done:
    return ret_addr;
}
#endif

/* virtual address to bus address                      */
/* XXX not currently used, but may be needed for Alpha */
static inline unsigned long 
get_bus_address(unsigned long address) 
{
    unsigned long ret_addr;

    if (address > VMALLOC_START)
        address = VMALLOC_VMADDR(address);

    /* make sure we don't seg fault on a bad address */
    ret_addr = get_kvirt_address(address);
    if (ret_addr) ret_addr = virt_to_bus( (void *) ret_addr);

    return ret_addr;
}

/* virtual address to physical page address */
unsigned long 
get_phys_address(unsigned long address) 
{
    unsigned long ret_addr;

    if (address > VMALLOC_START)
        address = VMALLOC_VMADDR(address);

    /* make sure we don't seg fault on a bad address */
    ret_addr = get_kvirt_address(address);
    if (ret_addr) ret_addr = __pa( (void *) ret_addr);

    return ret_addr;
}
/* increment physical page count and lock the pages */
static void nv_lock_pages(
    nv_state_t    *nv,
    void *address,
    unsigned long  num_pages)
{
    int i;
       
    for (i = 0; i < num_pages; i++)
    {
        unsigned long page = get_phys_address( (unsigned long) 
            (address + (i * PAGE_SIZE)));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	unsigned long map_nr = MAP_NR(__va(page));
#else
        struct page *map_nr = virt_to_page(__va(page));
#endif
        if (map_nr)
        {
            mem_map_reserve(map_nr);
            mem_map_inc_count(map_nr);
        }
    }
}

/* decrement physical page count. 
 * If the count is down to '1', unlock the physical page.
 * return '0' if all of the pages are unlocked, '1' otherwise.
 */
int nv_unlock_pages(
    nv_state_t   *nv,
    unsigned long address,
    unsigned long num_pages
)
{
    int i, its_all_good_flag = 0;

    for (i = 0; i < num_pages; i++)
    {
        unsigned long page = get_phys_address(address + (i * PAGE_SIZE));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	unsigned long map_nr = MAP_NR(__va(page));
#else
        struct page *map_nr = virt_to_page(__va(page));
#endif
        if (map_nr)
        {
            mem_map_dec_count(map_nr);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	    if (atomic_read(&mem_map[map_nr].count) == 1)
#else
            if (atomic_read(&map_nr->count) == 1)
#endif
                mem_map_unreserve(map_nr);
            else
                its_all_good_flag = 1;
        }
    }
    return its_all_good_flag;
}

void nv_lock_grab(
    unsigned long *pflags
)
{
    spin_lock_irqsave(&nv_lock, *pflags);
    nv_locked = pflags;		// make sure osAllocMem() knows not to sleep
}

void nv_lock_release(
    unsigned long flags
)
{
    nv_locked = (unsigned long *) 0;
    spin_unlock_irqrestore(&nv_lock, flags);
}


/*
 * Given a virtual address, fid the 'at' that owns it
 * Uses the physical address as the key.
 */
static nv_alloc_t *nv_find_alloc(
    nv_state_t    *nv,
    unsigned long  address,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;
    void *physaddr, *offset;
    unsigned int i;

    address = address & PAGE_MASK;
    physaddr = (void *) get_phys_address(address);

    if (physaddr == NULL)
        goto done;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nv->alloc_queue; at; *prev = at, at = at->next)
    {
        for (i = 0; i < at->num_pages; i++)
        {
            offset = (void *) get_phys_address((unsigned long) (at->kernel_address + (i << PAGE_SHIFT)));
            if ( (physaddr >= offset) && (physaddr < offset + PAGE_SIZE - 1) )
                return at;
        }

        /* check for AGP pages in the linear aperture */
        for (i = 0; i < at->num_pages; i++)
        {
            offset = at->physical_address + (i << PAGE_SHIFT);
            if ( (physaddr >= offset) && (physaddr < offset + PAGE_SIZE - 1) )
                return at;
        }
    }

done:

    // failure is not necessarily an error if the caller was just probing an address
    NV_DMSG(nv, "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}

/*
 * Given a virtual address, find the agp 'at' that owns it
 * Uses the process id as the key.
 * this is similar to the older nv_find_alloc, which is unsafe, due to pids 
 * and threading.. this one should be fine, since it's only used when setting 
 * up agp pages, which is all in one resman call.
 */
nv_alloc_t *nv_find_alloc_by_pid(
    nv_state_t    *nv,
    unsigned long  address,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    address = address & PAGE_MASK;
    *prev = (nv_alloc_t *) 0;
    for (at = nv->alloc_queue; at; *prev = at, at = at->next)
    {
        if ( (address >= at->vma->vm_start) && 
             (address < at->vma->vm_end) &&
             (at->process_id == current->pid) )
        {
            return at;
        }
    }

    // failure is not necessarily an error if the caller was just probing an address
    NV_DMSG(nv, "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}

/*
 * Given a virtual address, fid the agp 'at' that owns it
 * Uses the file pointer as the key.
 * Only used in unmap, when the pid may be different, 
 * and the ptes may already be zeroed out.
 */
nv_alloc_t *nv_find_alloc_by_file(
    nv_state_t    *nv,
    unsigned long  address,
    struct file *file,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    address = address & PAGE_MASK;
    *prev = (nv_alloc_t *) 0;
    for (at = nv->alloc_queue; at; *prev = at, at = at->next)
    {
        if ( (address >= at->vma->vm_start) && 
             (address < at->vma->vm_end) &&
             (at->file == file) )
        {
            return at;
        }
    }

    // failure is not necessarily an error if the caller was just probing an address
    NV_DMSG(nv, "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}

void *nv_find_kernel_mapping(
    nv_state_t    *nv,
    void          *address
)
{
    nv_alloc_t *at;
    at = nv_find_alloc(nv, (unsigned long) address, (nv_alloc_t **) 0);
    if (at)
        return at->kernel_address + ((unsigned long) address - at->vma->vm_start);

    return (void *) 0;
}


/* this is only called when the vmas are duplicated.
 * this appears to only happen when the process is cloned to create
 * a new process, and not when the process is threaded. 
 *
 * increment the usage count for the physical pages, so when this
 * clone unmaps the mappings, the pages are not deallocated under
 * the original process.
 */ 
void nv_vma_open(struct vm_area_struct *vma)
{
    nv_state_t *nv;

    nv = &nv_devices[NV_DEVICE_NUMBER(LINUX_VMA_DEV(vma))];
    NV_DMSG(nv, "vma_open [0x%lx-0x%lx]", vma->vm_start, vma->vm_end);

    if ( (LINUX_VMA_OFFS(vma) == NV_MMAP_ALLOCATION_OFFSET) ||
         (LINUX_VMA_OFFS(vma) == NV_MMAP_AGP_OFFSET) )
    {
        nv_alloc_t *at, *prev;
        at = nv_find_alloc(nv, vma->vm_start, &prev);
        if (at)
        {
            // increment physical page usage counts
            nv_lock_pages(nv, at->kernel_address, at->num_pages);
        }
    }

    MOD_INC_USE_COUNT;
}


void nv_vma_release(struct vm_area_struct *vma)
{
    nv_state_t *nv;

    nv = &nv_devices[NV_DEVICE_NUMBER(LINUX_VMA_DEV(vma))];
    NV_DMSG(nv, "vma_release [0x%lx-0x%lx]", vma->vm_start, vma->vm_end);

    MOD_DEC_USE_COUNT;
}


// assumes we are destroying entire map
void nv_vma_unmap(
    struct vm_area_struct *vma,
    unsigned long start,
    size_t len
)
{
    nv_alloc_t *at, *prev;
    nv_state_t *nv;

    nv = &nv_devices[NV_DEVICE_NUMBER(LINUX_VMA_DEV(vma))];

    NV_DMSG(nv, "unmap(0x%lx, 0x%x)", start, len);

    if ( (LINUX_VMA_OFFS(vma) == NV_MMAP_ALLOCATION_OFFSET) ||
         (LINUX_VMA_OFFS(vma) == NV_MMAP_AGP_OFFSET) )
    {
        at = nv_find_alloc_by_file(nv, vma->vm_start, vma->vm_file, &prev);
        if (at)
        {
            int num_pages = NV_PAGE_ALIGN(at->vma->vm_end - at->vma->vm_start);
            int unlocked = 0;

            // decrement physical page usage counts
            unlocked = nv_unlock_pages(nv, (unsigned long) at->kernel_address, 
                num_pages);

            // if nobody's using the pages anymore, free them
            if (!unlocked)
            {
                vfree(at->kernel_address);

                // take it out of the list
                if (prev == (nv_alloc_t *) 0)
                    nv->alloc_queue = at->next;
                else
                    prev->next = at->next;

                kfree(at);
            }
        } 
    }
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
struct vm_operations_struct nv_vm_ops = {
    nv_vma_open,
    nv_vma_release,  /* "close" */
    nv_vma_unmap,
    NULL,            /* protect */
    NULL,            /* sync */
    NULL,            /* advise */
    NULL,            // nv_vma_nopage,
    NULL,	     /* wppage */
    NULL,	     /* swapout */
    NULL	     /* swapin */
};
#else
struct vm_operations_struct nv_vm_ops = {
    open:     nv_vma_open,
    close:    nv_vma_release,
    unmap:    nv_vma_unmap,  
};
#endif

int init_module(void)
{
    nv_state_t *nv;
    int rc;

    // Register driver
    rc = register_chrdev(nv_major, "nvidia", &nv_fops);
    if (rc < 0) {
        NV_EMSG((nv_state_t *) 0,
                "init_module: register failed %d", rc);
        goto failed;
    }
    // if we requested dynamic major assignment, then 'rc' will have new number
    if (nv_major == 0)
        nv_major = rc;

    NV_DMSG((nv_state_t *) 0, "init_module: major number %d", nv_major);

    nv_probe_devices();

    // init all the bottom half structures
    for (nv = nv_devices; nv < nv_devices + NV_MAX_DEVICES; nv++)
    {
        nv->bh = &nv_bottom_halves[nv - nv_devices];
        nv->bh->routine = RmIsrBottomHalf;
        nv->bh->data = (void *) 0;
        nv->bh->sync = 0;
    }

    // Init the resource manager
    if ( ! RmInitRm())
    {
        NV_EMSG((nv_state_t *) 0, "RmInitRm() failed");
        rc = -EIO;
        goto failed;
    }

    return 0;

 failed:
    if (nv_major > 0)
        unregister_chrdev(nv_major, "nvidia");
    return rc;
}

void cleanup_module(void)
{
    int rc;
    nv_state_t *nv;

    NV_DMSG((nv_state_t *) 0, "cleanup_module");

    // Shutdown the resource manager
    RmShutdownRm();

    // make sure we freed up all the mappings
    // kernel should do this automagically before calling close
    for (nv = &nv_devices[0]; nv < nv_devices + NV_MAX_DEVICES; nv++)
        if (nv->alloc_queue)
            NV_EMSG(nv,
                    "still have vm que at cleanup_module(): 0x%x to 0x%x",
                    nv->alloc_queue->vma->vm_start,
                    nv->alloc_queue->vma->vm_end);

    if ((rc = unregister_chrdev(nv_major, "nvidia")) < 0)
        NV_EMSG((nv_state_t *) 0,
                "cleanup_module: unregister nv failed %d", rc);
}

static void nv_interrupts_disable(
    nv_state_t *nv
)
{
    nv->regs[NV_PMC_INTR_EN_0 / 4] = 0;
}

static void nv_interrupts_enable(
    nv_state_t *nv
)
{
    nv->regs[NV_PMC_INTR_EN_0 / 4] = 1;
}

// XXX: need to minimize mappings
//      the pci config space only tracks how much space is reserved.
//      It doesn't really know how big stuff is.
static int nv_map_device(
    nv_state_t *nv
)
{
    unsigned long size;

    // map the register space
    nv->regs = ioremap(nv->reg_physical_address, 16 * 1024 * 1024);

    // map the frame buffer
    // XXX should map measured size of fb, not fixed size
    //     unfortunately I don't know how to get the fb size.
    //     We have to know what it is before we call initHal()...
    //     current hack is to just keep mapping until we get something we can use
    //  NOTE:
    //     2.2.12 kernels seem to have a bug where they return the physical
    //     address if they can't get the memory, hence the check against physaddr

    size = 64 * 1024 * 1024;
    do {
        nv->fb = ioremap(nv->fb_physical_address, size);
        if ((nv->fb == (void *) 0) || (nv->fb == (void *) nv->fb_physical_address))
        {
            nv->fb = 0;
            size /= 2;
        }
    } while ((nv->fb == (void *) 0) && (size >= (4 * 1024 * 1024)));

    if (nv->fb == 0)
        return -ENXIO;

    nv->regs_u = (nv_phwreg_t) nv->regs;
    nv->fb_u = (nv_phwreg_t) nv->fb;
    return 0;
}

static void nv_unmap_device(
    nv_state_t *nv
)
{
    if (nv->regs)
        vfree(nv->regs);
    if (nv->fb)
        vfree(nv->fb);
    nv->regs = nv->fb = (void *) 0;
    nv->regs_u = nv->fb_u = (void *) 0;
}

/*
 * driver receives an interrupt
 *    if someone waiting, then hand it off.
 */

void nv_isr(
    int   irq,
    void *arg,
    struct pt_regs *regs
)
{
    nv_state_t *nv = (void *) arg;
    U032 need_to_run_bottom_half = 0;

    RmIsr(nv->device_number, &need_to_run_bottom_half);
    if (need_to_run_bottom_half)
    {
        nv->bh->data = nv->pdev;
        queue_task(nv->bh, &tq_immediate);
        mark_bh(IMMEDIATE_BH);
    }
}


//
// nv_open
//
// nv driver open entry point.  Sessions are created here.
//
int nv_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = (nv_state_t *) 0;
    nv_file_private_t *nvfp;
    int devnum;
    int rc, status;
    
    /* Grab a file private area and save 'nv' in file structure */
    nvfp = kmalloc(sizeof(nv_file_private_t), GFP_KERNEL);
    if ( ! nvfp)
        return -ENOMEM;
    memset(nvfp, 0, sizeof(*nvfp));
    file->private_data = nvfp;

    // for control device, just jump to its open routine
    // after setting up the private data
    if (NV_DEVICE_IS_CONTROL_DEVICE(inode->i_rdev))
        return nv_ctl_open(inode, file);

    // what device are we talking about?
    devnum = NV_DEVICE_NUMBER(inode->i_rdev);
    if (devnum >= NV_MAX_DEVICES)
    {
        rc = -ENODEV;
        goto failed;
    }

    MOD_INC_USE_COUNT;

    nv = &nv_devices[devnum];

    NV_DMSG(nv, "open by process %d (current is 0x%x)", current->pid, current);

    nvfp->nv = nv;

    /*
     * map the memory and allocate isr on first open
     */

    if ( ! (nv->flags & NV_FLAG_OPEN))
    {
        if (nv->device_id == 0)
        {
            NV_EMSG((nv_state_t *) 0, "open of nonexistent device %d", devnum);
            rc = -ENXIO;
            goto failed;
        }

        if ((rc = nv_map_device(nv)))
        {
            NV_EMSG(nv, "nv_map_device failed");
            goto failed;
        }

        if ( ! RmInitAdapter(nv))
        {
            NV_EMSG(nv, "RmInitAdapter failed");
            rc = -EIO;
            goto failed;
        }
        nv->flags |= NV_FLAG_POSTED;

        status = request_irq(nv->interrupt_line, nv_isr,
                             SA_INTERRUPT | SA_SHIRQ, "nvidia",
                             (void *) nv);
        if (status != 0)
        {
            if ( nv->interrupt_line && (status == -EBUSY) )
            {
                printk("NV: Tried to get irq %d, but another driver", 
                    (unsigned int) nv->interrupt_line);
                printk(" has it and is not sharing it.\n");
                printk("NV: you may want to verify that an audio driver");
                printk(" isn't using the irq\n");
            }
            NV_EMSG(nv, "isr request failed 0x%x", status);
            rc = -EIO;
            goto failed;
        }

        nv->event_queue = NULL;

        nv->flags |= NV_FLAG_OPEN;
    }

    nv->usage_count++;

    return 0;

 failed:
    MOD_DEC_USE_COUNT;
    if (nv)
        nv_unmap_device(nv);
    return rc;
}


//
// nv_close
//
// Master driver close entry point.
//

int nv_close(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = NV_GET_NV_FROM_FILEP(file);
    unsigned long lock_flags;

    // for control device, just jump to its open routine
    // after setting up the private data
    if (NV_DEVICE_IS_CONTROL_DEVICE(inode->i_rdev))
        return nv_ctl_close(inode, file);

    NV_DMSG(nv, "close");

    nv_lock_grab(&lock_flags);

    nv_free_all_unused_clients(nv, current->pid, (void *) file);

    if (--nv->usage_count == 0)
    {
        // turn off interrupts.
        // be careful to make sure any pending bottom half gets run
        //  or disabled before calling RmShutdownAdapter() since
        //  it will free up the pdev.  This is hard to see on single
        //  cpu systems, but easy on dual cpu :-)

        nv_interrupts_disable(nv);
        
        // give it a moment to allow any bottom half to run
        // have to release lock of course

        nv_lock_release(lock_flags);
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ/50);

        // regain lock and free the irq
        nv_lock_grab(&lock_flags);

        free_irq(nv->interrupt_line, (void *) nv);

        nv->bh->data = (void *) 0;

        RmShutdownAdapter(nv);

        (void) nv_unmap_device(nv);

        // make sure we have freed up all the mappings
        // kernel should do this automagically before calling close
        if (nv->alloc_queue)
            NV_EMSG(nv,
                    "still have vm que at nv_close(): 0x%x to 0x%x",
                    nv->alloc_queue->vma->vm_start,
                    nv->alloc_queue->vma->vm_end);

        // leave INIT flag alone so we don't reinit every time
        nv->flags &= ~(NV_FLAG_OPEN | NV_FLAG_WAITING);
    }

    // free up our per file private data
    if (file->private_data)
        kfree(file->private_data);
    file->private_data = (void *) 0;

    nv_lock_release(lock_flags);

    MOD_DEC_USE_COUNT;

    return 0;
}

int nv_mmap(
    struct file  *file,
    struct vm_area_struct *vma
)
{
    int pages;
    nv_alloc_t *at;
    nv_state_t *nv;
    nv = NV_GET_NV_FROM_FILEP(file);

    NV_DMSG(nv, "mmap([0x%lx-0x%lx] off=0x%lx", vma->vm_start, vma->vm_end, LINUX_VMA_OFFS(vma));

    // be a bit paranoid for now
    if ((NV_MASK_OFFSET(LINUX_VMA_OFFS(vma))) ||
        (NV_MASK_OFFSET(vma->vm_start)) ||
        (NV_MASK_OFFSET(vma->vm_end)))
    {
        return -ENXIO;
    }

    pages = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;

    // max is 64M
    if (pages > (64 * 1024 * 1024) / PAGE_SIZE)
        return -ENOMEM;

    // we have our own version to keep the module count right
    vma->vm_ops = &nv_vm_ops;
    
    /*
     * figure out the range and map it in
     */

    
    // NV reg space
    if ((LINUX_VMA_OFFS(vma) >= NV_MMAP_REG_OFFSET) &&
        ((LINUX_VMA_OFFS(vma) + (vma->vm_end - vma->vm_start)) <= NV_MMAP_FB_OFFSET))
    {
        if (remap_page_range(vma->vm_start,
                             (u32) (nv->reg_physical_address) + LINUX_VMA_OFFS(vma) - NV_MMAP_REG_OFFSET,
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        // mark it as IO so that we don't dump it on core dump
        vma->vm_flags |= VM_IO;
    }

    // NV fb space
    else if ((LINUX_VMA_OFFS(vma) >= NV_MMAP_FB_OFFSET) &&
             ((LINUX_VMA_OFFS(vma) + (vma->vm_end - vma->vm_start)) <= NV_MMAP_ALLOCATION_OFFSET))
    {
        if (remap_page_range(vma->vm_start,
                             (u32) (nv->fb_physical_address) + LINUX_VMA_OFFS(vma) - NV_MMAP_FB_OFFSET,
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        // mark it as IO so that we don't dump it on core dump
        vma->vm_flags |= VM_IO;
    }

    // Magic allocator 
    else if (LINUX_VMA_OFFS(vma) == NV_MMAP_ALLOCATION_OFFSET)
    {
        unsigned long page, pos, start;

        /*
         * save info about this range so that nopage can map it in
         */
        at = kmalloc(sizeof(nv_alloc_t) + (pages * sizeof(unsigned long)),
                     GFP_KERNEL);
        at->next = (void *) 0;
        at->vma = vma;
        at->kernel_address = vmalloc(pages << PAGE_SHIFT);
        at->physical_address = 
            (void *) get_phys_address((unsigned long) at->kernel_address);
        at->num_pages = pages;

        // increment physical page usage count
        nv_lock_pages(nv, at->kernel_address, pages);
        at->process_id = current->pid;
        at->file = file;
        at->next = nv->alloc_queue;
        nv->alloc_queue = at;

        pos = (unsigned long) at->kernel_address;
        start = vma->vm_start;
        while (pages--)
        {
            page = get_phys_address(pos);
            if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
              	return -EAGAIN;
           	start += PAGE_SIZE;
         	pos += PAGE_SIZE;
       	}

	}
    else if (LINUX_VMA_OFFS(vma) == NV_MMAP_AGP_OFFSET)
	{
        /*
         * save info about this range so that nopage can map it in
         */
        at = kmalloc(sizeof(nv_alloc_t) + (pages * sizeof(unsigned long)),
                     GFP_KERNEL);
        at->next = (void *) 0;
        at->vma = vma;
        at->kernel_address = vmalloc(pages << PAGE_SHIFT);
        at->physical_address = NULL;
        at->num_pages = pages;

        // increment physical page usage count
        nv_lock_pages(nv, at->kernel_address, pages);
        at->process_id = current->pid;
        at->file = file;
        at->next = nv->alloc_queue;
        nv->alloc_queue = at;

	} else
        return -EINVAL;

    vma->vm_file = file;

    /* just increment usage count, rather than calling vma_open */
    MOD_INC_USE_COUNT;

    return 0;
}

/*
 * Convert virtual (user or kernel) to "physical address"
 *   Here physical means the bus address, as we're just prepping
 *   to give it to dma
 */
int nv_vtop(
    nv_ioctl_memory_vtop_t *params,
    int                     copy_to_user
)
{
    void *v;
    int l;
    unsigned long p;		/* physical addr */
    u32 *physical_addresses;

    // page align everything

    v = NV_ALIGN_PTR_DOWN(params->buffer, PAGE_SIZE);
    l = NV_ALIGN_UP(params->byte_length + (params->buffer - v), PAGE_SIZE);

    for (physical_addresses = params->physical_addresses;
         l > 0;
         physical_addresses++)
    {
        // get the physical address and copy it out to caller's buffer
        p = (unsigned long) get_phys_address((unsigned long) v);
            
        NV_DMSG((nv_state_t *) 0, "vtop: 0x%p to p 0x%08x", v, (unsigned int) p);
        if (copy_to_user)
            put_user(p, physical_addresses);
        else
            *physical_addresses = p;
        l -= PAGE_SIZE;
        v += PAGE_SIZE;
    }

    return 0;
}

//
// post the event
// Called from osNotifyEvent()
//
// We already hold the spinlock when we get in here...
//
void nv_post_event(
    void     *file_void
)
{
    nv_state_t *nv;
    nv_file_private_t *nvfp;
    struct file *file = file_void;

    nvfp = file->private_data;
    nv = NV_GET_NV_FROM_FILEP(file);

    NV_DMSG(nv, "post");

    nvfp->any_fired_notifiers++;

#ifdef KERNEL_2_3
    wake_up_interruptible(nv->event_queue);
#else
    wake_up_interruptible(&nv->event_queue);
#endif
}

//
// post vblank specifically
// Called from osVblankNotify(), from VBlank()
//
void nv_post_vblank(
    nv_state_t *nv
)
{
    NV_DMSG(nv, "post");

    if (nv->waiting_for_vblank) 
        nv->vblank_notifier++;

#ifdef KERNEL_2_3
    wake_up_interruptible(nv->event_queue);
#else
    wake_up_interruptible(&nv->event_queue);
#endif
}

unsigned int nv_poll(
    struct file *file,
    poll_table *wait
)
{
    unsigned int mask = 0;
    nv_state_t *nv;
    nv_file_private_t *nvfp;
    unsigned long lock_flags;

    nvfp = file->private_data;
    nv = NV_GET_NV_FROM_FILEP(file);

    NV_DMSG(nv, "poll()");

    // add us to the list
#ifdef KERNEL_2_3
    poll_wait(file, nv->event_queue, wait);
#else
    poll_wait(file, &nv->event_queue, wait);
#endif

    nv_lock_grab(&lock_flags);

    // wake the user on any file-specific event, or a general vblank
    if (nvfp->any_fired_notifiers || nv->vblank_notifier)
    {
        if (nv->vblank_notifier)
        {
            // don't clear vblank_notifier until everyone's been notified
            nv->waiting_for_vblank--;
            if (nv->waiting_for_vblank == 0)
                nv->vblank_notifier = 0;
        }

        // tell user some have been posted
        mask |= POLLPRI;
        NV_DMSG(nv, "poll() live on entry!");
    } else {
        // keep track of how many people are waiting for vblank
        nv->waiting_for_vblank++;
    }

    nv_lock_release(lock_flags);

    return mask;
}

//
// nv_ioctl
//
// nv driver ioctl entry point.
//

#ifdef __KERNEL__
#define COPYIN(dst, src, bytes)     copy_from_user((dst), (src), (bytes))
#define COPYOUT(dst, src, bytes)   copy_to_user((dst), (src), (bytes))
#else
#define COPYIN(dst, src, bytes)     memcpy((dst), (src), (bytes)), 0
#define COPYOUT(dst, src, bytes)   memcpy((dst), (src), (bytes)), 0
#endif

#define DO_OR_DIE(n) { if (n) { status = -EFAULT; goto done; } }

/*
 * some ioctl's can only be done on actual device, others only on the control device
 */
#define CTL_DEVICE_ONLY(nv) { if ( ! ((nv)->flags & NV_FLAG_CONTROL)) { status = -EINVAL; goto done; } }

#define ACTUAL_DEVICE_ONLY(nv) { if ((nv)->flags & NV_FLAG_CONTROL) { status = -EINVAL; goto done; } }


/* todo:
   need ioctl to raise a thread priority that is not superuser
       set its priority to SCHED_FIFO which is simple
       priority scheduling w/ disabled timeslicing
*/

int nv_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int cmd,
    unsigned long i_arg)
{
    int size;
    int status = 0;
    nv_state_t *nv;
    void *arg = (void *) i_arg;

    nv = NV_GET_NV_FROM_FILEP(file);

    NV_DMSG(nv, "ioctl(%d, 0x%x)", _IOC_NR(cmd), (unsigned int) i_arg);

    size = _IOC_SIZE(cmd);

    switch (cmd)
    {
        // debug tool; zap the module use count so we can unload driver
        //             even if it is confused
        case NV_IOCTL_MODULE_RESET:
            atomic_set(&__this_module.uc.usecount, 1);
            break;

        // enable/disable interrupts
        // used by X driver to make card quiet while Xserver unnecessarily diddles
        // around in its PCI config space

        case NV_IOCTL_INTERRUPTS:
        {
            int state = i_arg;
            ACTUAL_DEVICE_ONLY(nv);
            NV_DMSG(nv, "%sabling interrupts", state ? "En" : "Dis");
            if (state)
                nv_interrupts_enable(nv);
            else
                nv_interrupts_disable(nv);
            break;
        }

        case NV_IOCTL_MEMORY_VTOP:
        {
            nv_ioctl_memory_vtop_t params;
            nv_ioctl_memory_vtop_t *userbuf = arg;

            ACTUAL_DEVICE_ONLY(nv);

            /* copy in the user command buffer */
            DO_OR_DIE(COPYIN(&params, userbuf, sizeof(params)));
            
            status = nv_vtop(&params, 1);

            break;
        }

        // pass out info about the card
        case NV_IOCTL_CARD_INFO:
        {
            nv_ioctl_card_info_t ci;
            nv_ioctl_card_info_t *userbuf = arg;
            nv_state_t *tnv;
            int i;

            CTL_DEVICE_ONLY(nv);

            for (i = 0, tnv = nv_devices; tnv < nv_devices + NV_MAX_DEVICES; tnv++, i++)
            {
                (void) memset(&ci, 0, sizeof(ci));
                if (tnv->device_id)
                {
                    ci.flags = NV_IOCTL_CARD_INFO_FLAG_PRESENT;
                    ci.bus = tnv->bus;
                    ci.slot = tnv->slot;
                    ci.vendor_id = tnv->vendor_id;
                    ci.device_id = tnv->device_id;
                    ci.interrupt_line = tnv->interrupt_line;
                }
                // copy the card info out to user
                // just copying 0's for nonexistent cards
                DO_OR_DIE(COPYOUT(userbuf + i, &ci, sizeof(ci)));
            }
            break;
        }
            

        default:
            status = nvExtEscape(file, nv, cmd, arg) ? 0 : -EINVAL;
            break;
    }

 done:
    return status;
}

//
// nv_ctl_open
//
// nv control driver open entry point.  Sessions are created here.
//
int nv_ctl_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = (nv_state_t *) 0;
    nv_file_private_t *nvfp;

    nv = &nv_ctl_device;
    nv->device_number = NV_CONTROL_DEVICE_NUMBER;

    // save the nv away in file->private_data
    nvfp = file->private_data;
    nvfp->nv = nv;

    MOD_INC_USE_COUNT;

    NV_DMSG(nv, "open");

    nv->flags |= NV_FLAG_OPEN + NV_FLAG_CONTROL;

    nv->usage_count++;

    return 0;
}


//
// nv_ctl_close
//
int nv_ctl_close(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv;
    nv = NV_GET_NV_FROM_FILEP(file);

    NV_DMSG(nv, "close");

    if (--nv->usage_count == 0)
        nv->flags = 0;

    nv_free_all_unused_clients(nv, current->pid, (void *) file);

    // free up our per file private data
    if (file->private_data)
        kfree(file->private_data);
    file->private_data = (void *) 0;

    MOD_DEC_USE_COUNT;
    return 0;
}

