 /***************************************************************************\
|*                                                                           *|
|*       Copyright 1993-1999 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NOTICE TO USER:   The source code  is copyrighted under  U.S. and     *|
|*     international laws.  Users and possessors of this source code are     *|
|*     hereby granted a nonexclusive,  royalty-free copyright license to     *|
|*     use this code in individual and commercial software.                  *|
|*                                                                           *|
|*     Any use of this source code must include,  in the user documenta-     *|
|*     tion and  internal comments to the code,  notices to the end user     *|
|*     as follows:                                                           *|
|*                                                                           *|
|*       Copyright 1993-1999 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NVIDIA, CORPORATION MAKES NO REPRESENTATION ABOUT THE SUITABILITY     *|
|*     OF  THIS SOURCE  CODE  FOR ANY PURPOSE.  IT IS  PROVIDED  "AS IS"     *|
|*     WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND.  NVIDIA, CORPOR-     *|
|*     ATION DISCLAIMS ALL WARRANTIES  WITH REGARD  TO THIS SOURCE CODE,     *|
|*     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, NONINFRINGE-     *|
|*     MENT,  AND FITNESS  FOR A PARTICULAR PURPOSE.   IN NO EVENT SHALL     *|
|*     NVIDIA, CORPORATION  BE LIABLE FOR ANY SPECIAL,  INDIRECT,  INCI-     *|
|*     DENTAL, OR CONSEQUENTIAL DAMAGES,  OR ANY DAMAGES  WHATSOEVER RE-     *|
|*     SULTING FROM LOSS OF USE,  DATA OR PROFITS,  WHETHER IN AN ACTION     *|
|*     OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,  ARISING OUT OF     *|
|*     OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOURCE CODE.     *|
|*                                                                           *|
|*     U.S. Government  End  Users.   This source code  is a "commercial     *|
|*     item,"  as that  term is  defined at  48 C.F.R. 2.101 (OCT 1995),     *|
|*     consisting  of "commercial  computer  software"  and  "commercial     *|
|*     computer  software  documentation,"  as such  terms  are  used in     *|
|*     48 C.F.R. 12.212 (SEPT 1995)  and is provided to the U.S. Govern-     *|
|*     ment only as  a commercial end item.   Consistent with  48 C.F.R.     *|
|*     12.212 and  48 C.F.R. 227.7202-1 through  227.7202-4 (JUNE 1995),     *|
|*     all U.S. Government End Users  acquire the source code  with only     *|
|*     those rights set forth herein.                                        *|
|*                                                                           *|
 \***************************************************************************/

#ifndef _NV_H_
#define _NV_H_

#include <nvtypes.h>

// Nvidia's reserved major device number
// Set this to 0 to request dynamic major number allocation.
#define NV_MAJOR_DEVICE_NUMBER 195

// most cards in a single system
#define NV_MAX_DEVICES 4

typedef struct nv_ioctl_memory_vtop
{
    void *buffer;
    int   byte_length;
    unsigned int *physical_addresses;
} nv_ioctl_memory_vtop_t;

#define NV_IOCTL_MAGIC      'F'
#define NV_IOCTL_MEMORY_VTOP    _IOWR(NV_IOCTL_MAGIC, 1, sizeof(void *))

typedef struct nv_ioctl_card_info
{
    int    flags;               // see below
    int    bus;			// bus number (PCI, AGP, etc)
    int    slot;                // card slot
    int    vendor_id;           // PCI vendor id
    int    device_id;
    int    interrupt_line;

} nv_ioctl_card_info_t;

#define NV_IOCTL_CARD_INFO_FLAG_PRESENT       0x0001

#define NV_IOCTL_CARD_INFO      _IOWR(NV_IOCTL_MAGIC, 2, sizeof(void *))


// allow interrupts to be turned off/on.
// hopefully just for debug or working around buggy Xservers :-)
// Specify 'arg' as a 0 (disable) or non-zero (enable)
#define NV_IOCTL_INTERRUPTS     _IOWR(NV_IOCTL_MAGIC, 3, sizeof(void *))


/* debug tool to force module count to 0 */
#define NV_IOCTL_MODULE_RESET  _IO(NV_IOCTL_MAGIC, 99)

/*
 * Power of 2 alignment.
 *    (Will give unexpected results if 'a' is not a power of 2.)
 */

#define NV_ALIGN_DOWN(v, gran) (((NvU32) (v)) & ~((gran) - 1))
#define NV_ALIGN_UP(v, gran)   ((((NvU32) (v)) + ((gran) - 1)) & ~((gran)-1))

#define NV_ALIGN_PTR_DOWN(p, gran) ((void *) NV_ALIGN_DOWN((p), (gran)))
#define NV_ALIGN_PTR_UP(p, gran) ((void *) NV_ALIGN_UP((p), (gran)))

#define NV_PAGE_ALIGN_DOWN(value) NV_ALIGN_DOWN((value), PAGE_SIZE)
#define NV_PAGE_ALIGN_UP(value)     NV_ALIGN_UP((value), PAGE_SIZE)


/* mmap(2) offsets */

#define NV_MMAP_REG_OFFSET              (0)
#define NV_MMAP_FB_OFFSET                (256 * 1024 * 1024)
// Magic mmap() offset for the memory allocator
#define NV_MMAP_ALLOCATION_OFFSET (1 * 1024 * 1024 * 1024)
// AGP indicator, intended to be or'd with magic allocator offset
#define NV_MMAP_AGP_OFFSET ((unsigned long)(2)*(1024 * 1024 * 1024))


#ifdef __KERNEL__

#include <linux/config.h>
#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif

#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 1, 0)
#error "This driver does not support pre-2.1 kernels"
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 0)
  #define KERNEL_2_1
#else
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 0)
    #define KERNEL_2_2
  #else
    #warning This driver is not officially supported on post-2.2 kernels
    #define KERNEL_2_3
  #endif  // < 2.3
#endif  // < 2.2

#include <linux/autoconf.h>

#include <os-agpconfig.h>

/*
 * ptr arithmetic convenience
 */

typedef union
{
    volatile V008 Reg008[1];
    volatile V016 Reg016[1];
    volatile V032 Reg032[1];
} nv_hwreg_t, * nv_phwreg_t;

/*
 * data structure to keep track of open clients for abnormal termination
 * cleanup.  See nv.c for discussion of cleanup issues
 * NOTE: this can't be kept in 'nv' since the nv current during root allocate
 *       is the control 'nv'.
 */

#define NV_MAX_CLIENTS 128         /* *system-wide* limit, not per device */
typedef struct {
    NvU32  client;
    int    pid;
	void  *file;
} nv_client_t;
extern nv_client_t nv_clients[NV_MAX_CLIENTS]; 

/*
 * An allocated bit of memory using NV_MEMORY_ALLOCATION_OFFSET
 *   looks like this in the driver
 */

typedef struct nv_alloc_s {
    struct nv_alloc_s *next;    
    struct vm_area_struct *vma;
    unsigned int   process_id;
    unsigned int   num_pages;
    void          *kernel_address;
    void          *physical_address;
    struct file   *file;
} nv_alloc_t;

/*
 * per device state
 */

typedef struct
{
    int    usage_count;
    int    flags;
    int    device_number;

    void  *pdev;		    // you know who

    nv_alloc_t *alloc_queue;

    // PCI config info
    int bus;
    int slot;
    int vendor_id;
    int device_id;
    
    // physical characteristics
    NvU32  reg_physical_address;
    NvU32  fb_physical_address;
    NvU32  interrupt_line;
    NvU32  interrupt_pin;

    // mapped addresses
    NvU32  *regs;
    NvU32  *fb;
    NvU32  *rom;
    nv_phwreg_t  regs_u, fb_u;	     // ptr arith convenience

    // bottom half interrupt handler info; per device
    struct tq_struct *bh;

    // queue for for NV's OS events
#ifdef KERNEL_2_3
    struct __wait_queue_head *event_queue;
#else
    struct wait_queue *event_queue;
#endif

    U032 vblank_notifier;
    U032 waiting_for_vblank;

    U032 agpgart;

} nv_state_t;


/*
 * flags
 */

#define NV_FLAG_OPEN    0x0001
#define NV_FLAG_POSTED  0x0002
#define NV_FLAG_WAITING 0x0004
#define NV_FLAG_CONTROL 0x0010

/*
 * poll(2) support for event notifiers
 *
 * A notifer can be turned into an "OS" event.
 * On UNIX, this means you can block using poll() while waiting for any
 * of your OS events to fire.
 *
 * When poll() on the NV file descriptor notices that you have one or more
 * events, it returns POLLPRI.  At that time it clears the count of fired
 * notifiers.
 *
 * Upon return from poll() just poll your notifiers.
 */

/*
 * Per-file private data structure.
 * We keep stuff here that is separate on a per-file basis.
 */

typedef struct {
    U032 any_fired_notifiers;
    nv_state_t *nv;
} nv_file_private_t;

#define NV_GET_NV_FROM_FILEP(filep) \
            ((nv_file_private_t *) (filep)->private_data)->nv

//
// where we hide our nv_state_t * in a pdev
// subject to change...
//
#define NV_HIDE_NV_IN_PDEV(pdev) \
                     ((nv_state_t *) (pdev)->DBmpDeviceExtension)


/*
 * driver internal interfaces
 */

// need a fake device number for control device; just to flag it for msgs
#define NV_CONTROL_DEVICE_NUMBER     100

unsigned long get_phys_address(unsigned long address);

nv_alloc_t *nv_find_alloc_by_pid(nv_state_t *, unsigned long, nv_alloc_t **);
int   nv_vtop(nv_ioctl_memory_vtop_t *params, int in_kernel);
void *nv_find_kernel_mapping(nv_state_t *, void *address);

void  nv_free_all_unused_clients(nv_state_t *nv, int pid, void *file);

extern unsigned long * volatile nv_locked;	// are we holding lock right now?
void nv_lock_grab(unsigned long *);
void nv_lock_release(unsigned long);
void nv_post_event(void *filep);
void nv_post_vblank(nv_state_t *nv);

#endif  /* __KERNEL__ */

/*
 * resman interfaces the driver has to know about
 */

int  RmInitRm(void);
int  RmShutdownRm(void);
int  RmIsr(U032 deviceReference, U032 *need_to_run_bottom_half);
void RmIsrBottomHalf(void *pdev_void);
#ifdef __KERNEL__
int  RmInitAdapter(nv_state_t *nv);
int  RmShutdownAdapter(nv_state_t *nv);
int  nvExtEscape(void *file, nv_state_t *nv, int cmd, void *pvParms);

#endif

#endif
